/*------------------------------------------------------------------------------
Name:      JdbcManagerCommonTable.java
Project:   xmlBlaster.org
Copyright: xmlBlaster.org, see xmlBlaster-LICENSE file
------------------------------------------------------------------------------*/

package org.xmlBlaster.util.queue.jdbc;

import org.jutils.log.LogChannel;
import org.xmlBlaster.util.XmlBlasterException;
import org.xmlBlaster.util.enum.ErrorCode;
import org.xmlBlaster.util.Global;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;

import java.util.Hashtable;
import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import org.xmlBlaster.util.queue.StorageId;
import org.xmlBlaster.util.queue.I_Entry;
import org.xmlBlaster.util.queue.I_EntryFactory;
import org.xmlBlaster.util.queue.StorageId;
import org.xmlBlaster.util.queue.ReturnDataHolder;
import org.xmlBlaster.util.queue.I_StorageProblemListener;
import org.xmlBlaster.util.queue.I_StorageProblemNotifier;

import org.xmlBlaster.util.Timestamp;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.TreeSet;
import java.util.WeakHashMap;
import java.util.StringTokenizer;
import java.util.Iterator;
import org.xmlBlaster.util.queue.QueuePluginManager;
import org.xmlBlaster.util.plugin.PluginInfo;

/**
 * Delegate class which takes care of SQL specific stuff for the JdbcQueuePlugin
 * class.
 * <p>
 * One instance of this is created by Global for each StorageId prefix,
 * so one instance for 'history', one for 'cb' etc.
 * </p>
 *
 *  to define:
 *             Postgres
 *  string
 *  longint
 *  int
 *  blob
 *
 * The tables needed for each JdbcManagerCommonTable instance are the following:
 * - nodesTableName  (defaults to 'NODES')
 * - queuesTableName (defaults to 'QUEUES')
 * - entriesTableName (defaults to 'ENTRIES')
 * The names of such tables are constituted by the tableNamePrefix (which defaults to 'XB') plus the 
 * nodesTableName
 * 
 * @author <a href="mailto:xmlBlaster@marcelruff.info">Marcel Ruff</a>
 * @author <a href='mailto:laghi@swissinfo.org'>Michele Laghi</a>
 *
 */
public class JdbcManagerCommonTable implements I_StorageProblemListener, I_StorageProblemNotifier {

   private static final String ME = "JdbcManagerCommonTable";
   private final Global glob;
   private final LogChannel log;
   private final JdbcConnectionPool pool;
   private final I_EntryFactory factory;
   private final WeakHashMap listener;
   private final static String DUMMY_VALUE = "A";
   private final TreeSet queues;

   private final String tableNamePrefix;
   // the names to be used
   private String stringTxt = null;
   private String longintTxt = null;
   private String intTxt = null;
   private String blobTxt = null;
   private String booleanTxt = null;
   private String tablesTxt = null;
   private String tableNameTxt = null;

   private boolean dbInitialized = false;
   private int maxStatementLength = 0;
   private int[] errCodes = null;
   private boolean isConnected = true;

   private static boolean first = true;
   private String nodesTableName = null;
   private String queuesTableName = null;
   private String entriesTableName = null;
   private boolean hasAdminPermission = true;

   private final int queueIncrement;

   /**
    * @param JdbcConnectionPool the pool to be used for the connections to
    *        the database. IMPORTANT: The pool must have been previously
    *        initialized.
    */
   public JdbcManagerCommonTable(JdbcConnectionPool pool, I_EntryFactory factory)
      throws XmlBlasterException {
      this.pool = pool;
      this.glob = this.pool.getGlobal();
      this.log = glob.getLog("jdbc");
      if (this.log.CALL) this.log.call(ME, "Constructor called");

      this.factory = factory;

      if (!this.pool.isInitialized())
         throw new XmlBlasterException(glob, ErrorCode.INTERNAL_UNKNOWN, ME, "constructor: the Connection pool is not properly initialized");

      // get the necessary metadata
      Connection conn = null;
      try {
         conn = this.pool.getConnection();
         this.maxStatementLength = conn.getMetaData().getMaxStatementLength();
         if (this.maxStatementLength < 1) {
            this.maxStatementLength = glob.getProperty().get("queue.persistent.maxStatementLength", 2048);
            if (first) {
               this.log.info(ME, "The maximum SQL statement length is not defined in JDBC meta data, we set it to " + this.maxStatementLength);
               first = false;
            }
         }

         if (!conn.getMetaData().supportsTransactions()) {
            String dbName = conn.getMetaData().getDatabaseProductName();
            this.log.error(ME, "the database '" + dbName + "' does not support transactions, unpredicted results may happen");
         }
      }
      catch (SQLException ex) {
         throw new XmlBlasterException(glob, ErrorCode.RESOURCE_CONFIGURATION, ME, "constructor: failed to get the metadata", ex);
      }
      finally {
         if (conn != null) this.pool.releaseConnection(conn);
      }

      Hashtable names = pool.getMapping();
      this.queues = new TreeSet();
      this.listener = new WeakHashMap();

      this.stringTxt = (String)names.get("string");
      if (this.stringTxt == null) this.stringTxt = "text";

      this.longintTxt = (String)names.get("longint");
      if (this.longintTxt == null) this.longintTxt = "bigint";

      this.intTxt = (String)names.get("int");
      if (this.intTxt == null) this.intTxt = "integer";

      this.booleanTxt = (String)names.get("boolean");
      if (this.booleanTxt == null) this.booleanTxt = "boolean";

      this.blobTxt = (String)names.get("blob");
      if (this.blobTxt == null) this.blobTxt = "bytea";

      this.tablesTxt = (String)names.get("tables");
      if (this.tablesTxt == null) this.tablesTxt = "pg_tables";
      this.tableNameTxt = (String)names.get("tablename");
      if (this.tableNameTxt == null) tableNameTxt = "tablename";

      this.tableNamePrefix = this.pool.getTableNamePrefix();
      this.queueIncrement = this.pool.getTableAllocationIncrement(); // 2

      String errorCodesTxt = (String)names.get("connectionErrorCodes");
      if (errorCodesTxt == null) errorCodesTxt = "1089:17002";

      StringTokenizer tokenizer = new StringTokenizer(errorCodesTxt, ":");
      this.errCodes = new int[tokenizer.countTokens()];
      int nmax = tokenizer.countTokens();
      for (int i=0; i < nmax; i++) {
         try {
            this.errCodes[i] = Integer.parseInt(tokenizer.nextToken().trim());
         }
         catch (Exception ex) {
            this.errCodes[i] = -1;
         }
      }
      if (this.log.DUMP) this.log.dump(ME, "Constructor: num of error codes: "  + nmax);

      this.nodesTableName = this.tableNamePrefix + 
                            (String)pool.getPluginProperties().getProperty("nodesTableName", "NODES");
      this.queuesTableName = this.tableNamePrefix + 
                            (String)pool.getPluginProperties().getProperty("queuesTableName", "QUEUES");
      this.entriesTableName = this.tableNamePrefix + 
                            (String)pool.getPluginProperties().getProperty("entriesTableName", "ENTRIES");
   }

  /**
   * Adds (registers) a listener for connection/disconnection events.
   */
   public boolean registerStorageProblemListener(I_StorageProblemListener entry) {
      synchronized (this.listener) {
         this.listener.put(entry, DUMMY_VALUE);  // use DUMMY_VALUE to support check in unregisterListener()
      }
      return true;
   }

  /**
   * Adds (registers) a listener for connection/disconnection events.
   * @return boolean true if the entry was successfully removed.
   */
   public boolean unRegisterStorageProblemListener(I_StorageProblemListener entry) {
      synchronized (this.listener) {
         return this.listener.remove(entry) != null;
      }
   }

   /**
    * @see I_StorageProblemListener#storageUnavailable(int)
    */
   public void storageUnavailable(int oldStatus) {
      if (this.log.CALL) this.log.call(ME, "disconnected invoked");
      this.isConnected = false;

      I_StorageProblemListener[] listenerArr = getStorageProblemListenerArr();
      for(int i=0; i<listenerArr.length; i++) {
         if (this.isConnected == true) break;
         I_StorageProblemListener singleListener = listenerArr[i];
         singleListener.storageUnavailable(oldStatus);
      }
   }

   /**
    * @see I_StorageProblemListener#storageAvailable(int)
    */
   public void storageAvailable(int oldStatus) {
      if (this.log.CALL) this.log.call(ME, "reconnected invoked");
      this.isConnected = true;
      //change this once this class implements I_StorageProblemNotifier
      if (oldStatus == I_StorageProblemListener.UNDEF) return;
      I_StorageProblemListener[] listenerArr = getStorageProblemListenerArr();
      for(int i=0; i<listenerArr.length; i++) {
         if (this.isConnected == false) break;
         I_StorageProblemListener singleListener = listenerArr[i];
         singleListener.storageAvailable(oldStatus);
      }
   }

   /**
    * @return A current snapshot of the connection listeners where we can work on (unsynchronized) and remove
    * listeners without danger
    */
   public I_StorageProblemListener[] getStorageProblemListenerArr() {
      synchronized (this.listener) {
         return (I_StorageProblemListener[])this.listener.keySet().toArray(new I_StorageProblemListener[this.listener.size()]);
      }
   }

   /**
    * @see #handleSQLException(String, SQLException, String)
    */
   protected final boolean handleSQLException(String location, SQLException ex) {
      return handleSQLException(location, ex, null);
   }

   /**
    * handles the SQLException. If it is a communication exception like the
    * connection has been broken it will inform the connection pool.
    * @param location where the exception occured.
    * @param ex the exception which has to be handled.
    * @param trace additional information to put in the logetLogId(tableName, storageId, "getEntries") logging trace.
    * @return boolean true if it was a communication exception
    * 
    */
   protected final boolean handleSQLException(String location, SQLException ex, String trace) {
      int err = ex.getErrorCode(), i=0;
      boolean ret = false;

      // postgres hack (since postgres does not return error codes other than 0)
      if (ex.getMessage().startsWith("FATAL"))
         ret = true;
      else {
         for (i=0; i < this.errCodes.length; i++) {
            if (this.errCodes[i] == err) break;
         }
         ret = (i < this.errCodes.length);
      }

      this.log.dump(ME, location + ": error code     : " + ex.getErrorCode());
      this.log.dump(ME, location + ": SQL state      : " + ex.getSQLState());
      if (trace != null && trace.length() > 0) {
         if (this.log.TRACE) this.log.trace(ME, location + ": additional info: " + trace);
      }

      if (ret) {
         this.log.error(ME, location + ": the connection to the DB has been lost. Going in polling modus");
         this.pool.setConnectionLost();
      }
      return ret;
   }

   /**
    * checks if all necessary tables exist. If a table does not exist and 'createTable' true, then the 
    * table is created.
    * @return boolean 'true' if the tables are all there after the invocation to this method, 'false' if at
    *         least one of the required tables is missing after the invocation to this method.
    */
   public final boolean tablesCheckAndSetup(boolean createTables)
      throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "tablesCheckAndSetup");

      if (!this.isConnected)
         throw new XmlBlasterException(glob, ErrorCode.RESOURCE_DB_UNAVAILABLE, ME, "tablesCheckAndSetup: the DB disconnected. Handling is currently not possible");

      boolean nodesTableExists = false;
      boolean queuesTableExists = false;
      boolean entriesTableExists = false;

      String req = "SELECT count(*) from " + tablesTxt + " where upper(" + tableNameTxt + ")='" + this.nodesTableName.toUpperCase() + "'";
      if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");

      PreparedQuery query = null;

      try {
         query = new PreparedQuery(pool, req, false, this.log, -1);
         query.rs.next();         
         int size = query.rs.getInt(1);
         if (size > 0) nodesTableExists = true;
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "number of '" + this.nodesTableName  + "' is " + size);

         req = "SELECT count(*) from " + tablesTxt + " where upper(" + tableNameTxt + ")='" + this.queuesTableName.toUpperCase() + "'";
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");
         query.inTransactionRequest(req /*, -1*/);
         query.rs.next();
         size = query.rs.getInt(1);
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "number of '" + this.queuesTableName  + "' is " + size);
         if (size > 0) queuesTableExists = true;

         req = "SELECT count(*) from " + tablesTxt + " where upper(" + tableNameTxt + ")='" + this.entriesTableName.toUpperCase() + "'";
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");
         query.inTransactionRequest(req /*, -1*/);
         query.rs.next();
         size = query.rs.getInt(1);
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "number of '" + this.entriesTableName  + "' is " + size);
         if (size > 0) entriesTableExists = true;

         if (!createTables) 
            return  nodesTableExists && queuesTableExists && entriesTableExists;


         if (!nodesTableExists) {
            log.info(getLogId(null, null, "tablesCheckAndSetup"), "adding table '" + this.nodesTableName + "' as the 'nodes' table");
            req = "CREATE TABLE " + this.nodesTableName.toUpperCase() + " (nodeId " + this.stringTxt + ", PRIMARY KEY (nodeId))";
            if (this.log.TRACE) 
               this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");
            update(req, query.conn);
         }

         if (!queuesTableExists) {
            log.info(getLogId(null, null, "tablesCheckAndSetup"), "adding table '" + this.queuesTableName + "' as the 'queues' table");
            req = "CREATE TABLE " + this.queuesTableName.toUpperCase() + " (queueName " + this.stringTxt +
                  ", nodeId " + this.stringTxt +
                  ", numOfBytes " + this.longintTxt +
                  ", numOfEntries " + this.longintTxt +
                  ", PRIMARY KEY (queueName, nodeId)" + 
                  ", FOREIGN KEY (nodeId) REFERENCES " + this.nodesTableName + " ON DELETE CASCADE)";
            if (this.log.TRACE) 
               this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");
            update(req, query.conn);
         }

         if (!entriesTableExists) {
            log.info(getLogId(null, null, "tablesCheckAndSetup"), "adding table '" + this.entriesTableName + "' as the 'entries' table");
            req = "CREATE TABLE " + this.entriesTableName.toUpperCase() + " (dataId " + this.longintTxt +
                  ", nodeId " + this.stringTxt +
                  ", queueName " + this.stringTxt +
                  ", prio " + this.intTxt +
                  ", flag " + this.stringTxt +
                  ", durable " + this.booleanTxt +
                  ", byteSize " + this.longintTxt +
                  ", blob " + this.blobTxt +
                  ", PRIMARY KEY (dataId, nodeId)" + 
                  ", FOREIGN KEY (queueName, nodeId) REFERENCES " + this.queuesTableName + " ON DELETE CASCADE)";
            if (this.log.TRACE) 
               this.log.trace(getLogId(null, null, "tablesCheckAndSetup"), "Request: '" + req + "'");
            update(req, query.conn);
         }

         return true;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(null, null, "tablesCheckAndSetup"), ex, "SQL request giving problems: " + req);
         throw ex;
      }
      finally {
         if (query !=null) query.close();
      }
   }

   /**
    * Adds a node to the DB.
    * @return boolean false if the node already existed, true otherwise.
    */
   public final boolean addNode(String nodeId) throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "addNode");
      String req = "SELECT count(*) from " + this.nodesTableName + " where nodeId='" + nodeId + "'";
      if (this.log.TRACE) this.log.trace(getLogId(null, nodeId, "addNode"), "Request: '" + req + "'");
      PreparedQuery query = null;
      PreparedStatement preStatement = null;
      try {
         query = new PreparedQuery(pool, req, false, this.log, -1);
         query.rs.next();         
         int size = query.rs.getInt(1);
         if (size > 0) return false;

         req = "INSERT INTO " + this.nodesTableName + " VALUES (?)";
         preStatement = query.conn.prepareStatement(req);
         preStatement.setString(1, nodeId);
         if (this.log.TRACE) this.log.trace(getLogId(null, nodeId, "addNode"), preStatement.toString());
         preStatement.executeUpdate();
         return true;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(null, nodeId, "addNode"), ex);
          throw ex;
      }
      finally {
         if (preStatement !=null) preStatement.close();
         if (query != null) query.close();
      }
   }

   /**
    * Removes a node from the DB.
    * @return boolean false if the node was not on the DB, true if it has been successfully removed.
    */
   public final boolean removeNode(String nodeId) throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "removeNode");
      String postfix =  this.nodesTableName + " where nodeId='" + nodeId + "'";
      String req = "SELECT count(*) from " + postfix;
      if (this.log.TRACE) this.log.trace(getLogId(null, nodeId, "removeNode"), "Request: '" + req + "'");
      PreparedQuery query = null;
      PreparedStatement preStatement = null;
      try {
         query = new PreparedQuery(pool, req, false, this.log, -1);
         query.rs.next();         
         int size = query.rs.getInt(1);
         if (size < 1) return false;

         // remove all queues associated with this node here ...

         req = "DELETE FROM " + postfix;
         update(req, query.conn);
         return true;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(null, nodeId, "removeNode"), ex);
          throw ex;
      }
      finally {
         if (preStatement !=null) preStatement.close();
         if (query != null) query.close();
      }
   }

   /**
    * Adds a node to the DB.
    * @return boolean false if the node already existed, true otherwise.
    */
   public final boolean addQueue(String queueName, String nodeId, long numOfBytes, long numOfEntries) throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "addQueue");
      String req = "SELECT count(*) from " + this.queuesTableName + " where (queueName='" + queueName + "' AND nodeId='" + nodeId + "')";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addQueue"), "Request: '" + req + "'");
      PreparedQuery query = null;
      PreparedStatement preStatement = null;
      try {
         query = new PreparedQuery(pool, req, false, this.log, -1);
         query.rs.next();         
         int size = query.rs.getInt(1);
         if (size > 0) return false;

         // remove all entries associated with this queue here ...

         req = "INSERT INTO " + this.queuesTableName + " VALUES (?,?,?,?)";
         preStatement = query.conn.prepareStatement(req);
         preStatement.setString(1, queueName);
         preStatement.setString(2, nodeId);
         preStatement.setLong(3, numOfBytes);
         preStatement.setLong(4, numOfEntries);

         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addQueue"), preStatement.toString());
         preStatement.executeUpdate();
         return true;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "addQueue"), ex);
          throw ex;
      }
      finally {
         if (preStatement !=null) preStatement.close();
         if (query != null) query.close();
      }

   }

   /**
    * Removes a node from the DB.
    * @return boolean false if the node was not on the DB, true if it has been successfully removed.
    */
   public final boolean removeQueue(String queueName, String nodeId) throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "removeQueue");
      String postfix = this.queuesTableName + " where (queueName='" + queueName + "' AND nodeId='" + nodeId + "')";
      String req = "SELECT count(*) from " + postfix;
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "removeQueue"), "Request: '" + req + "'");
      PreparedQuery query = null;
      PreparedStatement preStatement = null;
      try {
         query = new PreparedQuery(pool, req, false, this.log, -1);
         query.rs.next();         
         int size = query.rs.getInt(1);
         if (size < 1) return false;

         req = "DELETE FROM " + postfix;
         update(req, query.conn);
         return true;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "removeQueue"), ex);
          throw ex;
      }
      finally {
         if (preStatement !=null) preStatement.close();
         if (query != null) query.close();
      }
   }

   /**
    *
    * Adds a row to the specified queue table
    * @param tableName The name of the table on which to perform the operation
    * @param dataId The long specifying the unique Id of this entry in the queue
    * @param prio The priority of this entry
    * @param flag The flag which specifies which kind of object is stored in the blob
    * @param blob An object (must be Serializable) to store in the DB.
    *
    * @return true on success
    *
    * @throws SQLException if an error occured while adding the row
    * @throws XmlBlasterException if an error occured when trying to get a connection
    */
   public boolean addEntry(String queueName, String nodeId, I_Entry entry)
      throws SQLException, XmlBlasterException {

      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "addEntry"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addEntry"), "For entry '" + entry.getUniqueId() + "' currently not possible. No connection to the DB");
         return false;
      }

      long dataId = entry.getUniqueId();
      int prio = entry.getPriority();
      byte[] blob = this.factory.toBlob(entry);
      String typeName = entry.getEmbeddedType();
      boolean persistent = entry.isPersistent();
      long sizeInBytes = entry.getSizeInBytes();

      if (this.log.DUMP)
         this.log.dump(ME, "addition. dataId: " + dataId + ", prio: " + prio + ", typeName: " + typeName + ", byteSize in bytes: " + sizeInBytes);

      PreparedStatement preStatement = null;
      boolean ret = false;
      Connection conn = null;
      try {
         conn = this.pool.getConnection();
         String req = "INSERT INTO " + this.entriesTableName + " VALUES ( ?, ?, ?, ?, ?, ?, ?, ?)";
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addEntry"), req);

         preStatement = conn.prepareStatement(req);
         preStatement.setLong(1, dataId);
         preStatement.setString(2, nodeId);
         preStatement.setString(3, queueName);
         preStatement.setInt(4, prio);
         preStatement.setString(5, typeName);
         if (persistent == true) preStatement.setString(6, "T");
         else preStatement.setString(6, "F");
         preStatement.setLong(7, sizeInBytes);
         preStatement.setBytes(8, blob);

         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addEntry"), preStatement.toString());

         int num = preStatement.executeUpdate();
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "addEntry"), "Added " + num + " entries, entryId='" + entry.getUniqueId() + "'");
         ret = true;
      }
      catch (SQLException ex) {
         this.log.warn(getLogId(queueName, nodeId, "addEntry"), "Could not insert entry '" +
                  entry.getClass().getName() + "'-'" +  entry.getLogId() + "-" + entry.getUniqueId() + "': " + ex.toString());
         if (handleSQLException(getLogId(queueName, nodeId, "addEntry"), ex)) throw ex;

         // check if an entry with the same key already exists: if not, then an exception must be thrown
         try {
            String req = "SELECT count(*) from " + this.entriesTableName + " where (dataId='" + dataId + "' AND nodeId='" + nodeId + "')";
            ResultSet rs = conn.prepareStatement(req).executeQuery(req);
            rs.next();         
            int size = rs.getInt(1);
            if (size < 1) throw ex;
         }
         catch (SQLException ex1) {
            handleSQLException(getLogId(queueName, nodeId, "addEntry"), ex1);
            throw ex1;
         }
         ret = false;
      }
      finally {
         if (preStatement != null) preStatement.close();
         if (conn != null) this.pool.releaseConnection(conn);
      }

      return ret;
   }

   /**
    * Cleans up the specified queue. It deletes the queue from the 'queues' table which implicitly deletes
    * all entries in the 'entries' table associated to this queue.
    * @return the number of queues deleted (not the number of entries).
    */
   public final int cleanUp(String queueName, String nodeId) throws XmlBlasterException, SQLException {

      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "cleanUp"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "cleanUp"), "Currently not possible. No connection to the DB");
         return 0;
      }

      try {
         String req = "delete from " + this.queuesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "'";
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "cleanUp"), " request is '" + req + "'");
         return update(req);
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "deleteEntries"), ex);
         throw ex;
      }
   }


   public int wipeOutDB() throws XmlBlasterException {
      // retrieve all tables to delete
      String req = "SELECT " + this.tableNameTxt + " FROM " + this.tablesTxt + "";
      if (this.log.TRACE) this.log.trace(getLogId(null, null, "cleanUp"), "Request: '" + req + "'");
      PreparedQuery query = null;

      int count = 0;
      java.util.Vector vec = new java.util.Vector();
      try {
         query = new PreparedQuery(this.pool, req, this.log, -1);
         while (query.rs.next()) {
            String nameOfThisQueue = query.rs.getString(1).trim();
            if (nameOfThisQueue.toUpperCase().startsWith(this.tableNamePrefix)) vec.add(nameOfThisQueue);
         }

         for (int i=0; i < vec.size(); i++) {
            String name = (String)vec.elementAt(i);
            req = "DROP TABLE " + name + "";
            if (this.log.TRACE) this.log.trace(getLogId(null, null, "wipeOutDB"), "Request: '" + req + "'");
            try {
               this.update(req, query.conn);
               count++;
            }
            catch (Exception ex) {
               if (ex instanceof SQLException) {
                  if (handleSQLException(getLogId(null, null, "wipeOutDB"), (SQLException)ex))
                   throw new XmlBlasterException(glob, ErrorCode.RESOURCE_DB_UNKNOWN, getLogId(null, null, "wipeOutDB"), "SQLException when retrieving the list of tables", ex);
               }
               this.log.error(getLogId(null, null, "wipeOutDB"), "Could not delete queue '" + name + "'");
            }
         }
         return count;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(null, null, "wipeOutDB"), ex);
         throw new XmlBlasterException(glob, ErrorCode.RESOURCE_DB_UNKNOWN, getLogId(null, null, "wipeOutDB"), "SQLException when retrieving the list of tables", ex);
      }
      finally {
         try {
            if (query != null) query.close();
         }
         catch (Exception ex) {
            throw new XmlBlasterException(glob, ErrorCode.RESOURCE_DB_UNKNOWN, getLogId(null, null, "cleanUp"), "cleanUp: exception when closing the query", ex);
         }
      }
   }


   /**
    * Dumps the metadata to the log object. The log will write out this
    * information as an info. This means you don't need to have the switch
    * 'dump' set to true to see this information.
    */
   public void dumpMetaData() {
      this.pool.dumpMetaData();
   }


   /**
    * returns the amount of bytes currently in the specified queue
    * @param tableName the name of the table in which to count
    * @return the current amount of bytes used in the table.
    */
   public long getNumOfBytes(String queueName, String nodeId)
      throws SQLException, XmlBlasterException {

      if (!this.isConnected)
         throw new XmlBlasterException(glob, ErrorCode.RESOURCE_DB_UNAVAILABLE, getLogId(queueName, nodeId, "getNumOfBytes"), "The DB is disconnected. Handling queue '" + queueName + "' is currently not possible");
      Connection conn = null;
      PreparedStatement st = null;

      try {
         String req = "SELECT sum(byteSize) from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "'";
         conn = this.pool.getConnection();
         st = conn.prepareStatement(req);
         ResultSet rs = st.executeQuery();
         rs.next();
         return rs.getLong(1);
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getNumOfBytes"), ex);
         throw ex;
      }
      finally {
         if (st != null) st.close();
         if (conn != null) this.pool.releaseConnection(conn);
      }
   }


   /**
    * Sets up a table for the queue specified by this queue name.
    * If one already exists (i.e. if it recovers from a crash) its associated
    * table name is returned. If no such queue is found among the used table,
    * a new table is taken from the free tables. If no free tables are available,
    * a certain amount of such tables are created.
    */
   public final void setUp() throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(ME, "setUp");

      try {
         if (!this.dbInitialized) {
            if (this.log.TRACE) this.log.trace(getLogId(null, null, "setUp"), "Initializing the first time the pool");
            tablesCheckAndSetup(this.hasAdminPermission);
            this.dbInitialized = true;
         }
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(null, null, "setUp"), ex, "Table name giving Problems");
         throw ex;
      }
   }

   private final ArrayList processResultSet(ResultSet rs, StorageId storageId, int numOfEntries, long numOfBytes, boolean onlyId)
      throws SQLException, XmlBlasterException {

      if (this.log.CALL) this.log.call(getLogId(storageId.getStrippedId(), null, "processResultSet"), "Entering");

      ArrayList entries = new ArrayList();
      int count = 0;
      long amount = 0L;
      long currentAmount = 0L;

      while ( (rs.next()) && ((count < numOfEntries) || (numOfEntries < 0)) &&
         ((amount < numOfBytes) || (numOfBytes < 0))) {

         if(onlyId) {
            entries.add(new Long(rs.getLong(1)));
         }
         else {
            long dataId = rs.getLong(1);            // preStatement.setLong(1, dataId);
            /*String nodeId =*/ rs.getString(2);    // preStatement.setString(2, nodeId);
            /*String queueName =*/ rs.getString(3); // preStatement.setString(3, queueName);
            int prio = rs.getInt(4);                // preStatement.setInt(4, prio);
            String typeName = rs.getString(5);      // preStatement.setString(5, typeName);
//            boolean persistent = rs.getBoolean(4); // preStatement.setBoolean(4, persistent);
            //this only to make ORACLE happy since it does not support BOOLEAN
            String persistentAsChar = rs.getString(6);
            boolean persistent = false;
            if ("T".equalsIgnoreCase(persistentAsChar)) persistent = true;

            long sizeInBytes = rs.getLong(7);
            byte[] blob = rs.getBytes(8); // preStatement.setObject(5, blob);

            if ( (numOfBytes < 0) || (sizeInBytes+amount < numOfBytes) || (count == 0)) {
               if (this.log.DUMP)
                  this.log.dump(ME, "processResultSet: dataId: " + dataId + ", prio: " + prio + ", typeName: " + typeName + " persistent: " + persistent);
               entries.add(this.factory.createEntry(prio, dataId, typeName, persistent, blob, storageId));
               amount += sizeInBytes;
            }
         }
         count++;
      }

      while ((rs.next()) && ((count < numOfEntries) || (numOfEntries <0))) {
         if(onlyId) {
            entries.add(new Long(rs.getLong(1)));
         }
         else {
            long dataId = rs.getLong(1); // preStatement.setLong(1, dataId);
            int prio = rs.getInt(2); // preStatement.setInt(2, prio);
            String typeName = rs.getString(3); // preStatement.setString(3, typeName);
//            boolean persistent = rs.getBoolean(4); // preStatement.setBoolean(4, persistent);
            //this only to make ORACLE happy since it does not support BOOLEAN
            String persistentAsChar = rs.getString(4);
            boolean persistent = false;
            if ("T".equalsIgnoreCase(persistentAsChar)) persistent = true;

            long sizeInBytes = rs.getLong(5);
            byte[] blob = rs.getBytes(6); // preStatement.setObject(5, blob);

            if (this.log.DUMP)
               this.log.dump(ME, "processResultSet: dataId: " + dataId + ", prio: " + prio + ", typeName: " + typeName + " persistent: " + persistent);
            entries.add(this.factory.createEntry(prio, dataId, typeName, persistent, blob, storageId));
         }
         count++;
      }

      return entries;
   }


   /**
    * It accepts result sets with (long dataId, long size)
    * @param numOfBytes as input is the maximum number of bytes to process. As
    *        output it stores the number of bytes processed.
    * @param numOfEntries the maximum number of entries to process
    *
    */
   private final ReturnDataHolder processResultSetForDeleting(ResultSet rs, int numOfEntries, long numOfBytes)
      throws SQLException, XmlBlasterException {

      if (this.log.CALL) this.log.call(ME,"processResultSetForDeleting invoked");
      ReturnDataHolder ret = new ReturnDataHolder();
      long currentAmount = 0L;
      while ( (rs.next()) && ((ret.countEntries < numOfEntries) || (numOfEntries < 0)) &&
         ((ret.countBytes < numOfBytes) || (numOfBytes < 0))) {
         currentAmount = rs.getLong(2);
         if ( (numOfBytes < 0) || (ret.countBytes+currentAmount < numOfBytes) || (ret.countEntries == 0)) {
            ret.list.add(new Long(rs.getLong(1)));
            ret.countBytes += currentAmount;
            ret.countEntries++;
         }
      }
      return ret;
   }


   /**
    * Returns the pool associated to this object.
    * @return JdbcConnectionPool the pool managing the connections to the DB
    */
   public JdbcConnectionPool getPool() {
      return this.pool;
   }

   /**
    * A generic invocation for a simple update on the database. It will use
    * a normal statement (not a prepared statement),
    * @param request the request string
    * @param fetchSize the number of entries to fetch, i.e. to update
    */
   private final int update(String request)
      throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(getLogId(null, null, "update"), "Entering request=" + request);

      Connection conn = null;
      Statement  statement = null;
      int ret = 0;

      try {
         conn = pool.getConnection();
         statement = conn.createStatement();
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "update"), "Executing statement: " + request);
         ret = statement.executeUpdate(request);
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "update"), "Executed statement, number of changed entries=" + ret);
      }
      finally {
         if (statement !=null) statement.close();
         if (conn != null) this.pool.releaseConnection(conn);
      }

      return ret;
   }


   /**
    * This version makes an update but does no connection management, i.e. a
    * connection must be established outside this method and passed in the
    * argument list. The closing and cleaning up of the connection must also
    * be handled outside.
    *
    * @param request the string specifying which request to make
    * @param conn the connection to use for this request
    */
   private final int update(String request, Connection conn)
      throws SQLException {

      if (this.log.CALL) this.log.call(getLogId(null, null, "update"), "Request=" + request + " and connection " + conn);

      Statement  statement = null;
      int ret = 0;

      try {
         statement = conn.createStatement();
         ret = statement.executeUpdate(request);
         if (this.log.TRACE) this.log.trace(getLogId(null, null, "update"), "Executed statement, number of changed entries=" + ret);
      }

      finally {
         if (statement != null) statement.close();
      }

      return ret;
   }


   /**
    * deletes all transient messages
    */
   public int deleteAllTransient(String queueName, String nodeId) throws XmlBlasterException, SQLException {
      try {
         if (this.log.CALL) 
            this.log.call(getLogId(queueName, nodeId, "deleteAllTransient"), "deleteAllTransient");
         if (!this.isConnected) {
            if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "deleteAllTransient"), "Currently not possible. No connection to the DB");
            return 0;
         }

   //      String req = "delete from " + tableName + " where durable='false'";
         String req = "delete from " + this.entriesTableName + " where durable='F'";
         return update(req);
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "deleteAllTransient"), ex);
         throw ex;
      }
   }

   /**
    * The prefix is the initial part of the SQL update/query. Note that this
    * method can be used both for SELECT statements as for updates such as
    * DELETE or UPDATE.
    * An example of prefix:
    * "delete from tableName where dataId in(";
    */
   private ArrayList whereInStatement(String reqPrefix, long[] uniqueIds) {
      if (this.log.CALL) this.log.call(ME, "whereInStatement");
      final String reqPostfix = ")";
      boolean isFirst = true;
      int initialLength = reqPrefix.length() + reqPostfix.length() + 2;
      StringBuffer buf = new StringBuffer();
      int length = initialLength;
      int currentLength = 0;

      ArrayList ret = new ArrayList();

      for (int i=0; i<uniqueIds.length; i++) {
         String req = null;
         String entryId = Long.toString(uniqueIds[i]);
         currentLength = entryId.length();
         length += currentLength;
         if ((length>this.maxStatementLength) || (i == (uniqueIds.length-1))) { // then make the update

            if (i == (uniqueIds.length-1)) {
               if (!isFirst) buf.append(",");
               buf.append(entryId);
            }
            req = reqPrefix + buf.toString() + reqPostfix;
            ret.add(req);

            length = initialLength + currentLength;
            buf = new StringBuffer();

            isFirst = true;
         }

         if (!isFirst) {
            buf.append(",");
            length++;
         }
         else isFirst = false;
         buf.append(entryId);
      }

      return ret;
   }

   /**
    * Under the same transaction it gets and deletes all the entries which fit
    * into the constrains specified in the argument list.
    * @see org.xmlBlaster.util.queue.I_Queue#takeLowest(int, long, org.xmlBlaster.util.queue.I_QueueEntry, boolean)
    */
   public ReturnDataHolder getAndDeleteLowest(StorageId storageId, String nodeId, int numOfEntries, long numOfBytes,
      int maxPriority, long minUniqueId, boolean leaveOne) throws XmlBlasterException, SQLException {

      String queueName = storageId.getStrippedId();
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getAndDeleteLowest"), "Entering");
      ReturnDataHolder ret = new ReturnDataHolder();

      if (!this.isConnected) {
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "getAndDeleteLowest"), "Currently not possible. No connection to the DB");
         return ret;
      }

      PreparedQuery query = null;

      try {
         //inverse order here ...
         String req = "select * from " + this.entriesTableName + " WHERE queueName='" + queueName + "' AND nodeId='" + nodeId + "' ORDER BY prio ASC, dataid DESC";
         query = new PreparedQuery(pool, req, false, this.log, -1);

         // process the result set. Give only back what asked for (and only delete that)
         ResultSet rs = query.rs;

         int count = 0;
         long amount = 0L;
//         long maxNumOfBytes = numOfBytes;
         boolean doContinue = true;
         boolean stillEntriesInQueue = false;

         while ( (stillEntriesInQueue=rs.next()) && ((count < numOfEntries) || (numOfEntries < 0)) &&
            (doContinue)) {

            long dataId = rs.getLong(1);            // preStatement.setLong(1, dataId);
            /*String nodeId =*/ rs.getString(2);    // preStatement.setString(2, nodeId);
            /*String queueName =*/ rs.getString(3); // preStatement.setString(3, queueName);
            int prio = rs.getInt(4);                // preStatement.setInt(4, prio);
            String typeName = rs.getString(5);      // preStatement.setString(5, typeName);
//            boolean persistent = rs.getBoolean(4); // preStatement.setBoolean(4, persistent);
            //this only to make ORACLE happy since it does not support BOOLEAN
            String persistentAsChar = rs.getString(6);
            boolean persistent = false;
            if ("T".equalsIgnoreCase(persistentAsChar)) persistent = true;

            long sizeInBytes = rs.getLong(7);
            byte[] blob = rs.getBytes(8); // preStatement.setObject(5, blob);

            // check if allowed or already outside the range ...
            if (((numOfBytes<0)||(sizeInBytes+amount<numOfBytes)||(count==0)) &&
               ((prio<maxPriority) || ((prio==maxPriority)&&(dataId>minUniqueId)) )) {
               if (this.log.DUMP) this.log.dump(getLogId(queueName, nodeId, "getAndDeleteLowest"), "dataId: " + dataId + ", prio: " + prio + ", typeName: " + typeName + " persistent: " + persistent);
               ret.list.add(this.factory.createEntry(prio, dataId, typeName, persistent, blob, storageId));
               amount += sizeInBytes;
               if (amount > numOfBytes) doContinue = false;
               if (numOfBytes < 0) doContinue = true;
            }
            else doContinue = false;
            count++;
         }

         ret.countBytes = amount;
         ret.countEntries = count;

         // prepare for deleting (we don't use deleteEntries since we want
         // to use the same transaction (and the same connection)
         if (leaveOne) {
            // leave at least one entry
            if (stillEntriesInQueue) stillEntriesInQueue = rs.next();
            if ((!stillEntriesInQueue) && (ret.list.size()>0))
               ret.list.remove(ret.list.size()-1);
         }
         //first strip the unique ids:
         long[] uniqueIds = new long[ret.list.size()];
         for (int i=0; i < uniqueIds.length; i++)
            uniqueIds[i] = ((I_Entry)ret.list.get(i)).getUniqueId();

         String reqPrefix = "delete from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND dataId in(";
         ArrayList reqList = this.whereInStatement(reqPrefix, uniqueIds);
         for (int i=0; i < reqList.size(); i++) {
            req = (String)reqList.get(i);
            if (this.log.TRACE)
               this.log.trace(getLogId(queueName, nodeId, "getAndDeleteLowest"), "'delete from " + req + "'");
            update(req, query.conn);
         }

         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getAndDeleteLowest"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * Deletes the entries specified by the entries array. Since all entries
    * are deleted at the same time, in case of an exception the result
    * is uncertain (it depends on what the used database will do). In case the
    * statement would be too long for the DB, the request is split into several 
    * statements which are invoked separately.
    *
    * @param   tableName the name of the table on which to delete the entries
    * @param   uniqueIds the array containing all the uniqueId for the entries to delete.
    */
   public int deleteEntries(String queueName, String nodeId, long[] uniqueIds)
      throws XmlBlasterException, SQLException {
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "deleteEntries"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "deleteEntries"), "Currently not possible. No connection to the DB");
         return 0;
      }

      try {
         int count = 0;
         String reqPrefix = "delete from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND dataId in(";
         ArrayList reqList = this.whereInStatement(reqPrefix, uniqueIds);

         for (int i=0; i < reqList.size(); i++) {
            String req = (String)reqList.get(i);
            if (this.log.TRACE)
               this.log.trace(getLogId(queueName, nodeId, "deleteEntries"), "'delete from " + req + "'");

            count +=  update(req);
         }
         return count;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "deleteEntries"), ex);
         throw ex;
      }
   }


   /**
    * Deletes number of entries specified by the argument list. If less than
    * that amount of elements is given, then all entries will be deleted.
    * If you specify -1 as the number of entries, then all entries will be
    * deleted.
    *
    * @param tableName the name of the table on which to act.
    * @param numOfEntries the number of entries to be deleted.
    * @param amount the maximum amount of bytes to remove. Note that if no entries
    *        fit into this size, the first entry is taken anyway (to avoid deadlocks)
    */
   public ReturnDataHolder deleteFirstEntries(String queueName, String nodeId, long numOfEntries, long amount)
      throws XmlBlasterException, SQLException {
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "deleteFirstEntries called"), "Entering");

      ReturnDataHolder ret = new ReturnDataHolder();
      if (!this.isConnected) {
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "deleteFirstEntries"), "Currently not possible. No connection to the DB");
         return ret;
      }

      if (numOfEntries > Integer.MAX_VALUE)
         throw new XmlBlasterException(glob, ErrorCode.INTERNAL_ILLEGALARGUMENT, getLogId(queueName, nodeId, "deleteFirstEntries"),
               "The number of entries=" + numOfEntries + " to be deleted is too big for this system");

      PreparedQuery query = null;

//      int count = 0;

  //    ArrayList list = null;
      try {
         String req = "select dataId,byteSize from " + this.entriesTableName + " WHERE queueName='" + queueName + "' AND nodeId='" + nodeId + "' ORDER BY prio DESC, dataid ASC";
         query = new PreparedQuery(pool, req, false, this.log, -1);
         // I only want the uniqueId (dataId)
         if (numOfEntries >= Integer.MAX_VALUE)
            throw new XmlBlasterException(glob, ErrorCode.INTERNAL_ILLEGALARGUMENT, getLogId(queueName, nodeId, "deleteFirstEntries"),
                     "The number of entries=" + numOfEntries + " to delete exceeds the maximum allowed byteSize");

         ret = processResultSetForDeleting(query.rs, (int)numOfEntries, amount);

         int nmax = ret.list.size();
         if (nmax < 1) return ret;

         long[] uniqueIds = new long[nmax];
         for (int i=0; i < nmax; i++) uniqueIds[i] = ((Long)ret.list.get(i)).longValue();

         ArrayList reqList = whereInStatement("delete from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND dataId in(", uniqueIds);
         ret.countEntries = 0L;

         // everything in the same transaction (just in case)
         for (int i=0; i < reqList.size(); i++) {
            req = (String)reqList.get(i);
            if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "deleteFirstEntries"), "'" + req + "'");

            ret.countEntries +=  update(req, query.conn);
         }

         return ret;

      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "deleteFirstEntries"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * gets the first numOfEntries of the queue which have the priority in the
    * range specified by prioMin and prioMax (inclusive).
    * If there are not so many entries in the queue, all elements in the queue
    * are returned.
    *
    * @param storageId the storageId of the queue/storage from which to retrieve the information.
    * @param nodeId the nodeId of the actual cluster 
    * @param numOfEntries the maximum number of elements to retrieve. If negative there is no constriction.
    * @param numOfBytes the maximum number of bytes to retrieve. If negative, there is no constriction.
    * @param minPrio the minimum priority to retreive (inclusive). 
    * @param maxPrio the maximum priority to retrieve (inclusive).
    *
    */
   public ArrayList getEntriesByPriority(StorageId storageId, String nodeId, int numOfEntries, long numOfBytes, int minPrio, int maxPrio)
      throws XmlBlasterException, SQLException {

      String queueName = storageId.getStrippedId();

      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getEntriesByPriority"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesByPriority"), "Currently not possible. No connection to the DB");
         return new ArrayList();
      }

      String req = "SELECT * from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND prio >= " + minPrio + " and prio <= " + maxPrio +
            " ORDER BY prio DESC, dataid ASC";
                                                                 
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesByPriority"), "Request: '" + req + "'");

      PreparedQuery query =null;
      try {
         query = new PreparedQuery(pool, req, this.log, numOfEntries);
         ArrayList ret = processResultSet(query.rs, storageId, numOfEntries, numOfBytes, false);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesByPriority"), "Found " + ret.size() + " entries");
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getEntriesByPriority"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * gets the first numOfEntries of the queue which have the same priority.
    * If there are not so many entries in the queue, all elements in the queue
    * are returned.
    *
    * @param numOfEntries the maximum number of elements to retrieve
    *
    */
   public ArrayList getEntriesBySamePriority( StorageId storageId, String nodeId, int numOfEntries, long numOfBytes)
      throws XmlBlasterException, SQLException {
      String queueName = storageId.getStrippedId();
      // 65 ms (for 10000 msg)
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getEntriesBySamePriority"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE)
            this.log.trace(getLogId(queueName, nodeId, "getEntriesBySamePriority"), "Currently not possible. No connection to the DB");
         return new ArrayList();
      }

      String req = null;
      req = "SELECT * from " + this.entriesTableName + " where queueName='" + queueName + "' and nodeId='" + nodeId + 
            "' and prio=(select max(prio) from " + this.entriesTableName + " where queueName='" + queueName + 
            "' AND nodeId='" + nodeId + "')  ORDER BY dataid ASC";

      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesBySamePriority"), "Request: '" + req + "'");

      PreparedQuery query = null;
      try {
         query = new PreparedQuery(pool, req, this.log, numOfEntries);
         ArrayList ret = processResultSet(query.rs, storageId, numOfEntries, numOfBytes, false);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesBySamePriority"), "Found " + ret.size() + " entries");
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getEntriesBySamePriority"), ex);
         throw ex;
      }
      finally {
         if(query != null) query.close();
      }
   }


   /**
    * gets the first numOfEntries of the queue.
    * If there are not so many entries in the queue, all elements in the queue
    * are returned.
    *
    * @param numOfEntries the maximum number of elements to retrieve
    *
    */
   public ArrayList getEntries(StorageId storageId, String nodeId, int numOfEntries, long numOfBytes)
      throws XmlBlasterException, SQLException {
      String queueName = storageId.getStrippedId();
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getEntries"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Currently not possible. No connection to the DB");
         return new ArrayList();
      }

      String req = "SELECT * from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' ORDER BY prio DESC, dataid ASC";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Request: '" + req + "' wanted limits: numOfEntries="+numOfEntries+" numOfBytes="+numOfBytes);
      PreparedQuery query = null;
      try {
         query = new PreparedQuery(pool, req, this.log, numOfEntries);
         ArrayList ret = processResultSet(query.rs, storageId, numOfEntries, numOfBytes, false);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Found " + ret.size() + " entries. Wanted limits: numOfEntries="+numOfEntries+" numOfBytes="+numOfBytes);
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getEntries"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }



   /**
    * gets the first numOfEntries of the queue until the limitEntry is reached.
    *
    * @param numOfEntries the maximum number of elements to retrieve
    */
   public ArrayList getEntriesWithLimit(StorageId storageId, String nodeId, I_Entry limitEntry)
      throws XmlBlasterException, SQLException {
      String queueName = storageId.getStrippedId();
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getEntriesWithLimit"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesWithLimit"), "Currently not possible. No connection to the DB");
         return new ArrayList();
      }

      int limitPrio = limitEntry.getPriority();
      long limitId = limitEntry.getUniqueId();
      String req = "SELECT * from " + this.entriesTableName + " WHERE queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND (prio > " + limitPrio + " OR (prio = " + limitPrio + " AND dataId < "  + limitId + ") ) ORDER BY prio DESC, dataid ASC";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesWithLimit"), "Request: '" + req + "'");
      PreparedQuery query = null;
      try {
         query = new PreparedQuery(pool, req, this.log, -1);
         ArrayList ret = processResultSet(query.rs, storageId, -1, -1L, false);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntriesWithLimit"), "Found " + ret.size() + " entries");
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getEntriesWithLimit"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * gets all the entries which have the dataid specified in the argument list.
    * If the list is empty or null, an empty ArrayList object is returned.
    */
   public ArrayList getEntries(StorageId storageId, String nodeId, long[] dataids)
      throws XmlBlasterException, SQLException {
      String queueName = storageId.getStrippedId();
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getEntries"), "Entering");
      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Currently not possible. No connection to the DB");
         return new ArrayList();
      }

      String req = null;
      if ((dataids == null) || (dataids.length < 1)) return new ArrayList();

      req = "SELECT * FROM " + this.entriesTableName + " WHERE queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND dataid in (";

      ArrayList requests = this.whereInStatement(req, dataids);

      PreparedQuery query = null;
      try {
         req = (String)requests.get(0);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Request: '" + req + "'");
         query = new PreparedQuery(pool, req, this.log, -1);

         ArrayList ret = processResultSet(query.rs, storageId, -1, -1L, false);

         for (int i=1; i < requests.size(); i++) {
            req = (String)requests.get(i);
            if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Request: '" + req + "'");
            query.inTransactionRequest(req /*, -1 */);
            ret.addAll(processResultSet(query.rs, storageId, -1, -1L, false));
         }
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getEntries"), "Found " + ret.size() + " entries");
         return ret;

      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getEntries"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }

   }


   /**
    * gets the real number of entries. that is it really makes a call to the DB to find out
    * how big the size is.
    */
   public final long getNumOfEntries(String queueName, String nodeId)
      throws SQLException, XmlBlasterException {
      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfEntries"), "Currently not possible. No connection to the DB");
         return 0L;
      }

      String req = "select count(*) from " + this.entriesTableName + " WHERE queueName='" + queueName + "' AND nodeId='" + nodeId + "'";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfEntries"), "Request: '" + req + "'");
      PreparedQuery query = null;

      try {
         query = new PreparedQuery(pool, req, this.log, -1);
         query.rs.next();
         long ret = query.rs.getLong(1);
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfEntries"), "Num=" + ret);
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getNumOfEntries"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * gets the real number of persistent entries, that is it really makes a call to the DB to find out
    * how big the size is.
    */
   public final long getNumOfPersistents(String queueName, String nodeId)
      throws SQLException, XmlBlasterException {

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfPersistents"), "Currently not possible. No connection to the DB");
         return 0L;
      }

      String req = "select count(*) from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND durable='T'";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfPersistent"), "Request: '" + req + "'");
      PreparedQuery query = null;
      try {
         query = new PreparedQuery(pool, req, this.log, -1);
         query.rs.next();
         long ret = query.rs.getLong(1);
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getNumOfPersistents"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * gets the real size of persistent entries, that is it really makes a call to the DB to find out
    * how big the size is.
    */
   public final long getSizeOfPersistents(String queueName, String nodeId)
      throws SQLException, XmlBlasterException {
      if (this.log.CALL) this.log.call(getLogId(queueName, nodeId, "getSizeOfPersistent"), "Entering");

      if (!this.isConnected) {
         if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getSizeOfPersistent"), "Currently not possible. No connection to the DB");
         return 0L;
      }

      String req = "select sum(bytesize) from " + this.entriesTableName + " where queueName='" + queueName + "' AND nodeId='" + nodeId + "' AND durable='T'";
      if (this.log.TRACE) this.log.trace(getLogId(queueName, nodeId, "getNumOfPersistents"), "Request: '" + req + "'");
      PreparedQuery query = null;
      try {
         query = new PreparedQuery(pool, req, this.log, -1);
         query.rs.next();
         long ret = query.rs.getLong(1);
         return ret;
      }
      catch (SQLException ex) {
         handleSQLException(getLogId(queueName, nodeId, "getNumOfPersistents"), ex);
         throw ex;
      }
      finally {
         if (query != null) query.close();
      }
   }


   /**
    * @param queueName or null
    * @param nodeId or null
    * @return A nice location string for logging (instead of plain ME)
    */
   private final String getLogId(String queueName, String nodeId, String methodName) {
      StringBuffer sb = new StringBuffer(200);
      sb.append(ME);
      if (this.tableNamePrefix != null) {
         sb.append("-").append(this.tableNamePrefix);
      }
      if (nodeId != null) {
         sb.append("-").append(nodeId);
      }
      if (queueName != null) {
         sb.append("-").append(queueName);
      }
      if (methodName != null) {
         sb.append("-").append(methodName).append("()");
      }
      return sb.toString();
   }

   /**
    * This main method can be used to delete all tables on the db which start
    * with a certain prefix. It is useful to cleanup the entire DB.
    * 
    * IMPORTANT: caution must be used to avoid to delete the wrong data.
    * <pre>
    * java org.xmlBlaster.util.queue.jdbc.JdbcManagerCommonTable
    * and enter XMLBLASTER or TEST
    * </pre>
    */
   public static void main(String[] args) {
      Global glob = Global.instance();
      glob.init(args);

      try {
         QueuePluginManager pluginManager = new QueuePluginManager(glob);
         PluginInfo pluginInfo = new PluginInfo(glob, pluginManager, "JDBC", "1.0");
         
         JdbcManagerCommonTable manager = glob.getJdbcQueueManagerCommonTable(pluginInfo);
         manager.setUp();
         manager.addNode("Frodo");
         manager.addQueue("queue1", "Frodo", 1000, 3000);
         manager.addQueue("queue2", "Frodo", 2000, 2000);
         manager.addQueue("queue3", "Frodo", 3000, 1000);

         manager.addNode("Fritz");
         manager.addQueue("queue1", "Fritz", 1000, 3000);
         manager.addQueue("queue2", "Fritz", 2000, 2000);
         manager.addQueue("queue3", "Fritz", 3000, 1000);

         manager.removeQueue("queue1", "Fritz");
         manager.removeQueue("queue2", "Fritz");
         manager.removeQueue("queue3", "Fritz");
         manager.removeNode("Fritz");
      }
      catch (Exception ex) {
         System.err.println("Main" + ex.toString());
      }

   }

}
